package net.libertacasa.pubsh.web;

import java.security.Principal;
import java.time.Instant;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import org.keycloak.KeycloakPrincipal;
import org.keycloak.KeycloakSecurityContext;
import org.keycloak.adapters.KeycloakDeployment;
import org.keycloak.adapters.RefreshableKeycloakSecurityContext;
import org.keycloak.adapters.springsecurity.token.KeycloakAuthenticationToken;
import org.keycloak.representations.IDToken;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.SpringApplication;
import org.springframework.boot.autoconfigure.SpringBootApplication;
import org.springframework.http.MediaType;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Controller;
import org.springframework.ui.Model;
import org.springframework.util.MultiValueMap;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.servlet.mvc.support.RedirectAttributes;
import com.github.dockerjava.api.model.Container;
import com.github.dockerjava.api.model.Image;
import com.github.kagkarlsson.scheduler.Scheduler;

@SpringBootApplication
@Controller
public class WebApplication {
	
	//private final CounterService counter;
	private static Scheduler scheduler;
	
	@Autowired
	public void Scheduler(Scheduler scheduler) {
		WebApplication.scheduler = scheduler;
	}

    static ArrayList<String> availableOs = new ArrayList<String>();
    
	public static void main(String[] args) {
		SpringApplication.run(WebApplication.class, args);
		
		//to-do: move this to some config file or database
        availableOs.add("archlinux");
        availableOs.add("opensuse-leap");
        availableOs.add("opensuse-tumbleweed");
        availableOs.add("ubuntu");		
	}

	@GetMapping("/")
	public String index() {
		return "redirect:/portal";
		
	}
	
	@GetMapping("/admin")
	public String portal(Model model) throws pubshError {
		KeycloakAuthenticationToken authentication = (KeycloakAuthenticationToken)
        SecurityContextHolder.getContext().getAuthentication();
		Principal principal = (Principal) authentication.getPrincipal();
		String username="";
		String email="";
		String attribute01="";
		
      if (! (principal instanceof KeycloakPrincipal)) {
    	  System.out.println("Fatal: received logon without Keycloak principal.");
    	  throw new pubshError("Invalid request.");
      }      
      if (principal instanceof KeycloakPrincipal) {
          KeycloakPrincipal<?> kPrincipal = (KeycloakPrincipal<?>) principal;
          IDToken token = kPrincipal.getKeycloakSecurityContext().getIdToken();
          Map<String, Object> customClaims = token.getOtherClaims();
          if (! customClaims.containsKey("username")) {
        	  System.out.println("Fatal: received logon without username.");        	  
        	  throw new pubshError("Invalid request.");
          }          
          if (customClaims.containsKey("email")) {
              email = String.valueOf(customClaims.get("email"));
          }        
          if (customClaims.containsKey("username")) {
        	  username = String.valueOf(customClaims.get("username"));
          }
          if (customClaims.containsKey("attribute01")) {
        	  attribute01 = String.valueOf(customClaims.get("attribute01"));
          }
          model.addAttribute("username", username);
          model.addAttribute("principalid", principal.getName());
          model.addAttribute("email", email);
          model.addAttribute("attribute01", attribute01);          
      }
      
      List<Image> images = Docker.getImages(null);
      List<Container> containers = Docker.getContainers(null);
           
      model.addAttribute("docker_images", images);
      model.addAttribute("docker_containers", containers);
      
      model.addAttribute("availableOs", availableOs);
      
      return("admin");
	}
	
	@GetMapping("/portal")
	public String user(Model model) throws pubshError {
		KeycloakAuthenticationToken authentication = (KeycloakAuthenticationToken)
        SecurityContextHolder.getContext().getAuthentication();
		Principal principal = (Principal) authentication.getPrincipal();
		String username="";
		String email="";
		String attribute01="";
      if (! (principal instanceof KeycloakPrincipal)) {
    	  System.out.println("Fatal: received logon without Keycloak principal.");
    	  throw new pubshError("Invalid request.");
      }
      if (principal instanceof KeycloakPrincipal) {
          KeycloakPrincipal<?> kPrincipal = (KeycloakPrincipal<?>) principal;
          IDToken token = kPrincipal.getKeycloakSecurityContext().getIdToken();
          Map<String, Object> customClaims = token.getOtherClaims();     
          if (! customClaims.containsKey("username")) {
        	  System.out.println("Fatal: received logon without username.");        	  
        	  throw new pubshError("Invalid request.");
          }
          if (customClaims.containsKey("email")) {
              email = String.valueOf(customClaims.get("email"));
          }        
          if (customClaims.containsKey("username")) {
        	  username = String.valueOf(customClaims.get("username"));
          }
          if (customClaims.containsKey("attribute01")) {
        	  attribute01 = String.valueOf(customClaims.get("attribute01"));
          }
          model.addAttribute("username", username);
          model.addAttribute("principalid", principal.getName());
          model.addAttribute("email", email);
          model.addAttribute("attribute01", attribute01);          
      }
      
      List<Image> images = Docker.getImages(username);
      List<Container> containers = Docker.getContainers(username);
           
      model.addAttribute("docker_images", images);
      model.addAttribute("docker_containers", containers);
      
      model.addAttribute("availableOs", availableOs); 
      
      return("portal");
	}	
	
	@DeleteMapping("/frontend/container/delete/{id}")
	public static String deleteContainer(@PathVariable String id, HttpServletRequest request, RedirectAttributes redirectAttributes) {
	 // [Start] This block should move to a logging method. It's only job is to print user details to the console.
	     KeycloakAuthenticationToken principal = (KeycloakAuthenticationToken) request.getUserPrincipal();
	     String username= null;
	     String userid = principal.getName();
	     IDToken token = principal.getAccount().getKeycloakSecurityContext().getIdToken();		
	     Map<String, Object> customClaims = token.getOtherClaims();
	     username = String.valueOf(customClaims.get("username"));
 	 // [End]
	        
		System.out.printf("Deletion triggered for ID %s by %s (%s)\n", id, userid, username);
		
		Docker.deleteContainer(id);
			
		return("redirect:/admin");
	}
	
	@DeleteMapping("/frontend/image/delete/{id}")
	public static String deleteImage(@PathVariable String id, HttpServletRequest request, RedirectAttributes redirectAttributes) {
	// [Start] This block should move to a logging method. It's only job is to print user details to the console.
	     KeycloakAuthenticationToken principal = (KeycloakAuthenticationToken) request.getUserPrincipal();
	     String username= null;
	     String userid = principal.getName();
	     IDToken token = principal.getAccount().getKeycloakSecurityContext().getIdToken();		
	     Map<String, Object> customClaims = token.getOtherClaims();
	     username = String.valueOf(customClaims.get("username"));
	 // [End]
	     
	     System.out.printf("Image deletion triggered for ID %s by %s (%s)\n", id, userid, username);
	     
	     try {
	    	 Docker.deleteImage(id);
	    	 String returnmessage =  "Success - removed image with ID " + id + "!";
	    	 redirectAttributes.addFlashAttribute("message", returnmessage);
	     } catch (com.github.dockerjava.api.exception.ConflictException exception) {
	    	 String returnmessage = "Error - Image is still being used by a container.";
	    	 redirectAttributes.addFlashAttribute("message", returnmessage);
	     }
	     		
	     return("redirect:/admin");
	}
	
	@PostMapping(path="/frontend/container/add",consumes=MediaType.APPLICATION_FORM_URLENCODED_VALUE)
	public static String addContainer(@RequestBody MultiValueMap<String, String> body, HttpServletRequest request, RedirectAttributes redirectAttributes) {
	     KeycloakAuthenticationToken principal = (KeycloakAuthenticationToken) request.getUserPrincipal();
	     String username = null;
	     String userid = principal.getName();
	     IDToken token = principal.getAccount().getKeycloakSecurityContext().getIdToken();		
	     Map<String, Object> customClaims = token.getOtherClaims();
	     username = String.valueOf(customClaims.get("username"));
	     String os = (String) body.getFirst("osChoice");
	     
	     System.out.printf("New container with OS %s requested by %s (%s)\n", os, userid, username);
	     
	     Random rand = new Random();
	     Integer randomInt = rand.nextInt(9999999-1111);
	     Integer count = randomInt;
	     
	     try {
	    	 String imageid = Docker.buildImage(username, os, count);
	    	 String returnmessage =  "Success - built image with ID " + imageid + " and tag sh" + count + "!";
	    	 redirectAttributes.addFlashAttribute("message", returnmessage);
	     } catch (Exception exception) {
	    	 redirectAttributes.addFlashAttribute("message", "Success!");
	    	 String returnmessage =  "Error - failed to build image :-(";
	    	 redirectAttributes.addFlashAttribute("message", returnmessage);
	    	 System.out.println(exception);
	     }
	     
	     //Docker.createContainer(imageid);
	     
	     return("redirect:/portal");
	}
	
	@PostMapping(path="/frontend/shell/add",consumes=MediaType.APPLICATION_FORM_URLENCODED_VALUE)
	public static String addShell(@RequestBody MultiValueMap<String, String> body, HttpServletRequest request, RedirectAttributes redirectAttributes) {
	     KeycloakAuthenticationToken principal = (KeycloakAuthenticationToken) request.getUserPrincipal();
	     String username = null;
	     String userid = principal.getName();
	     IDToken token = principal.getAccount().getKeycloakSecurityContext().getIdToken();		
	     Map<String, Object> customClaims = token.getOtherClaims();
	     username = String.valueOf(customClaims.get("username"));
	     String os = (String) body.getFirst("osChoice");
	     Integer expiry = (Integer) Integer.parseInt(body.getFirst("expiry"));
	     
	     System.out.printf("New shell with OS %s requested by %s (%s)\n", os, userid, username);
	     
	     Random rand = new Random();
	     Integer randomInt = rand.nextInt(9999999-1111);
	     Integer count = randomInt;
	      
	     try {
	    	 String containerid = Docker.createShell(username, os, count);
	    	 String instanceid = username + "&" + containerid;
	    	 scheduler.schedule(SchedulerBean.shellRemovalTask().instance(instanceid), Instant.now().plusSeconds(expiry));
	    	 String returnmessage =  "Success - spawned shell " + username + "_" + os + "_" + count + " - internal ID: " + containerid;
	    	 redirectAttributes.addFlashAttribute("message", returnmessage);
	     } catch (Exception exception) {
	    	 redirectAttributes.addFlashAttribute("message", "Success!");
	    	 String returnmessage =  "Error - failed to build image :-(";
	    	 redirectAttributes.addFlashAttribute("message", returnmessage);
	    	 System.out.println(exception);
	    	 //exception.printStackTrace();
	     }
	     
	     return("redirect:/portal");
	}
	
	@DeleteMapping("/frontend/shell/delete/{id}")
	public static String deleteShell(@PathVariable String id, HttpServletRequest request, RedirectAttributes redirectAttributes) {
	 // [Start] This block should move to a logging method. It's only job is to print user details to the console.
	     KeycloakAuthenticationToken principal = (KeycloakAuthenticationToken) request.getUserPrincipal();
	     String username= null;
	     String userid = principal.getName();
	     IDToken token = principal.getAccount().getKeycloakSecurityContext().getIdToken();		
	     Map<String, Object> customClaims = token.getOtherClaims();
	     username = String.valueOf(customClaims.get("username"));
 	 // [End]
	        
		System.out.printf("Deletion triggered for ID %s by %s (%s)\n", id, userid, username);
		
		Docker.deleteShell(username, id);
			
		return("redirect:/portal");
	}	
	
	@GetMapping(path = "/logout")
	public String logout(HttpServletRequest request) throws ServletException {
	  request.logout();
	  keycloakSessionLogout(request);
	  return "logout";
	 }
	
   private void keycloakSessionLogout(HttpServletRequest request){
      RefreshableKeycloakSecurityContext c = getKeycloakSecurityContext(request);
      KeycloakDeployment d = c.getDeployment();
      c.logout(d);
   }

   private RefreshableKeycloakSecurityContext getKeycloakSecurityContext(HttpServletRequest request){
	  return (RefreshableKeycloakSecurityContext) request.getAttribute(KeycloakSecurityContext.class.getName());
   }
   
   public class pubshError extends Exception
   {
     private static final long serialVersionUID = 1L;

	public pubshError(String message)
     {
       super(message);
     }
   }   
   
}
